#include <stdlib.h>
#include <math.h>
#include "elemtools.h"
#include "matrix.h"

/**
   function generates outer normal %vector to required element end
   
   @param enid - id of end node
   @param n - %vector of the outer normal
   
   JK, 8.3.2018
*/
void bar_normal_vectors (long enid,vector &n)
{
  if (enid==0){
    n[0]=-1.0;
  }
  if (enid==1){
    n[0]=1.0;
  }
}

/**
   function generates outer normal %vector to required element edge
   
   @param edgeid - id of edge
   @param x,y - arrays of node coordinates
   @param n - %vector of the outer normal
   
   JK, 3.3.2018
*/
void triangle_normal_vectors (long edgeid,vector &x,vector &y,vector &n)
{
  double norm,ss,alpha;
  vector t1(2),t2(2);
  
  if (edgeid==0){
    // direction vector of the required edge
    t1[0]=x[1]-x[0];
    t1[1]=y[1]-y[0];
    //  auxiliary vector
    t2[0]=x[0]-x[2];
    t2[1]=y[0]-y[2];
  }
  if (edgeid==1){
    // direction vector of the required edge
    t1[0]=x[2]-x[1];
    t1[1]=y[2]-y[1];
    //  auxiliary vector
    t2[0]=x[1]-x[0];
    t2[1]=y[1]-y[0];
  }
  if (edgeid==2){
    // direction vector of the required edge
    t1[0]=x[0]-x[2];
    t1[1]=y[0]-y[2];
    //  auxiliary vector
    t2[0]=x[2]-x[1];
    t2[1]=y[2]-y[1];
  }
  
  //  length of the direction vector
  norm = t1[0]*t1[0]+t1[1]*t1[1];
  if (norm<1.0e-10){
    print_err("nonpositive length of the direction vector",__FILE__,__LINE__,__func__);
    abort ();
  }

  scprd (t1,t2,ss);
  alpha = (0.0-ss)/norm;
  
  n[0]=t2[0]-alpha*t1[0];
  n[1]=t2[1]-alpha*t1[1];

  norm = sqrt(n[0]*n[0]+n[1]*n[1]);
  n[0]/=norm;
  n[1]/=norm;

}


/**
   function computes volume of an tetrahedral finite element
   
   @param eid - element id (it is used only in the case of nonpositive volume)
   @param x,y,z - arrays of node coordinates
   
   JK, 3. 3. 2018
*/
double tetrahedra_volume (long eid,vector &x,vector &y,vector &z)
{
  double vol,det;
  
  det = det3d (x.a, y.a, z.a);
  
  if(det <= 0.0){
    print_err("nonpositive volume of an element %d (the volume is %e)", __FILE__, __LINE__, __func__, eid+1, det);
    abort();
  }
  
  vol = det/6.0;
  
  return vol;
}


/**
   function generates outer normal %vector to required element surface
   
   @param surfid - id of surface
   @param x,y,z - arrays of node coordinates
   @param n - %vector of the outer normal
   
   JK, 3.3.2018
*/
void tetrahedra_normal_vectors (long surfid,vector &x,vector &y,vector &z,vector &n)
{
  double norm;
  vector t1(3),t2(3);
  
  if (surfid==0){
    // first in-plane vector
    t1[0]=x[2]-x[3];
    t1[1]=y[2]-y[3];
    t1[2]=z[2]-z[3];
    //  second in-plane vector
    t2[0]=x[1]-x[3];
    t2[1]=y[1]-y[3];
    t2[2]=z[1]-z[3];
  }
  if (surfid==1){
    // first in-plane vector
    t1[0]=x[0]-x[3];
    t1[1]=y[0]-y[3];
    t1[2]=z[0]-z[3];
    //  second in-plane vector
    t2[0]=x[2]-x[3];
    t2[1]=y[2]-y[3];
    t2[2]=z[2]-z[3];
  }
  if (surfid==2){
    // first in-plane vector
    t1[0]=x[1]-x[3];
    t1[1]=y[1]-y[3];
    t1[2]=z[1]-z[3];
    //  second in-plane vector
    t2[0]=x[0]-x[3];
    t2[1]=y[0]-y[3];
    t2[2]=z[0]-z[3];
  }
  if (surfid==3){
    // first in-plane vector
    t1[0]=x[1]-x[0];
    t1[1]=y[1]-y[0];
    t1[2]=z[1]-z[0];
    //  second in-plane vector
    t2[0]=x[2]-x[0];
    t2[1]=y[2]-y[0];
    t2[2]=z[2]-z[0];
  }
  
  //  outer normal vector to the first element surface
  n[0]=t1[1]*t2[2]-t1[2]*t2[1];
  n[1]=t1[2]*t2[0]-t1[0]*t2[2];
  n[2]=t1[0]*t2[1]-t1[1]*t2[0];
  
  //  length of the normal vector
  norm = sqrt(n[0]*n[0]+n[1]*n[1]+n[2]*n[2]);
  if (norm<1.0e-10){
    print_err("nonpositive length of the normal vector",__FILE__,__LINE__,__func__);
    abort ();
  }
  
  n[0]/=norm;
  n[1]/=norm;
  n[2]/=norm;
}

/**
   function computes areas of the tetrahedron surfaces
   
   @param surfid - id of surface
   @param x,y,z - arrays of node coordinates
   
   JK, 4.3.2018
*/
double tetrahedra_surface_areas (long surfid,vector &x,vector &y,vector &z)
{
  double area,l,l1,l2,l3;
  
  if (surfid==0){
    l1=sqrt((x[2]-x[1])*(x[2]-x[1]) + (y[2]-y[1])*(y[2]-y[1]) + (z[2]-z[1])*(z[2]-z[1]));
    l2=sqrt((x[3]-x[2])*(x[3]-x[2]) + (y[3]-y[2])*(y[3]-y[2]) + (z[3]-z[2])*(z[3]-z[2]));
    l3=sqrt((x[1]-x[3])*(x[1]-x[3]) + (y[1]-y[3])*(y[1]-y[3]) + (z[1]-z[3])*(z[1]-z[3]));
  }
  if (surfid==1){
    l1=sqrt((x[0]-x[2])*(x[0]-x[2]) + (y[0]-y[2])*(y[0]-y[2]) + (z[0]-z[2])*(z[0]-z[2]));
    l2=sqrt((x[0]-x[3])*(x[0]-x[3]) + (y[0]-y[3])*(y[0]-y[3]) + (z[0]-z[3])*(z[0]-z[3]));
    l3=sqrt((x[3]-x[2])*(x[3]-x[2]) + (y[3]-y[2])*(y[3]-y[2]) + (z[3]-z[2])*(z[3]-z[2]));
  }
  if (surfid==2){
    l1=sqrt((x[1]-x[0])*(x[1]-x[0]) + (y[1]-y[0])*(y[1]-y[0]) + (z[1]-z[0])*(z[1]-z[0]));
    l2=sqrt((x[3]-x[1])*(x[3]-x[1]) + (y[3]-y[1])*(y[3]-y[1]) + (z[3]-z[1])*(z[3]-z[1]));
    l3=sqrt((x[3]-x[0])*(x[3]-x[0]) + (y[3]-y[0])*(y[3]-y[0]) + (z[3]-z[0])*(z[3]-z[0]));
  }
  if (surfid==3){
    l1=sqrt((x[1]-x[0])*(x[1]-x[0]) + (y[1]-y[0])*(y[1]-y[0]) + (z[1]-z[0])*(z[1]-z[0]));
    l2=sqrt((x[2]-x[1])*(x[2]-x[1]) + (y[2]-y[1])*(y[2]-y[1]) + (z[2]-z[1])*(z[2]-z[1]));
    l3=sqrt((x[2]-x[0])*(x[2]-x[0]) + (y[2]-y[0])*(y[2]-y[0]) + (z[2]-z[0])*(z[2]-z[0]));
  }
  
  // Heron's formula for the triangle area
  l=(l1+l2+l3)/2.0;
  area=sqrt(l*(l-l1)*(l-l2)*(l-l3));
  
  return area;
}
